#ifdef AVR
#include <Arduino.h>
#else
#include <cstdint>
#include <cstring>
#endif

#include "ccsds_spacepacket.h"


namespace CCSDS 
{

SpacePacket::SpacePacket(void *p_SpContext, TSpCallback *p_SpCallback)
{
  mu32_Index = 0;
  mu8_PacketVersionNumber = 0;
  me_PacketType = SpacePacket::TM;
  mb_SecHdrFlag = false;
  mu16_APID = 0x000;
  me_SequenceFlags = SpacePacket::Unsegmented;
  mu16_PacketSequenceCount = 0;
  mu16_PacketDataLength = 0;
  mu16_SyncErrorCount = 0;
  mb_Overflow = 0;

  mp_SpContext = p_SpContext;
  mp_SpCallback = p_SpCallback;
}


void SpacePacket::setCallback(void *p_SpContext, TSpCallback *p_SpCallback)
{
  mp_SpContext = p_SpContext;
  mp_SpCallback = p_SpCallback;
}



uint32_t SpacePacket::create(uint8_t *pu8_Buffer, const uint32_t u32_BufferSize,
                 const PacketType e_PacketType, const SequenceFlags e_SequenceFlags, const uint16_t u16_APID, const uint16_t u16_SequenceCount, 
                 const uint8_t *pu8_PacketData, const uint16_t u16_PacketDataLength)
{
  return create(pu8_Buffer, u32_BufferSize, e_PacketType, e_SequenceFlags, u16_APID, u16_SequenceCount, 
                 NULL, 0, pu8_PacketData, u16_PacketDataLength);
}



uint32_t SpacePacket::create(uint8_t *pu8_Buffer, const uint32_t u32_BufferSize,
                 const PacketType e_PacketType, const SequenceFlags e_SequenceFlags, const uint16_t u16_APID, const uint16_t u16_SequenceCount, 
                 const uint8_t *pu8_SecondaryHeaderData, const uint16_t u16_SecondaryHeaderLength,    
                 const uint8_t *pu8_PacketData, const uint16_t u16_PacketDataLength)
 {
   if(!pu8_Buffer || (u32_BufferSize<7) 
     || (u32_BufferSize<6UL+u16_SecondaryHeaderLength+u16_PacketDataLength) 
     || (u16_PacketDataLength==0) || !pu8_PacketData)
     return 0;

   if(u16_SecondaryHeaderLength>0 && !pu8_SecondaryHeaderData)
     return 0;

   if((uint32_t)u16_SecondaryHeaderLength+(uint32_t)u16_PacketDataLength-1>0x3fff)
     return 0;  

   // create primary header
   _create_primary_header(pu8_Buffer, e_PacketType, e_SequenceFlags, u16_APID, u16_SequenceCount, 
       u16_SecondaryHeaderLength?true:false, (uint32_t)u16_SecondaryHeaderLength+(uint32_t)u16_PacketDataLength);

   if(u16_SecondaryHeaderLength>0)
     memcpy((char*)&pu8_Buffer[PrimaryHdrSize], (const char*)pu8_SecondaryHeaderData, u16_SecondaryHeaderLength);
   memcpy((char*)&pu8_Buffer[PrimaryHdrSize+u16_SecondaryHeaderLength], pu8_PacketData, u16_PacketDataLength);

   return PrimaryHdrSize+u16_SecondaryHeaderLength+u16_PacketDataLength;
 }                



uint32_t SpacePacket::createIdle(uint8_t *pu8_Buffer, const uint32_t u32_BufferSize,
                 const uint16_t u16_SequenceCount, const uint16_t u16_TargetPacketSize)
{
   if(!pu8_Buffer || (u32_BufferSize<u16_TargetPacketSize)
     || (u16_TargetPacketSize<PrimaryHdrSize+1))
     return 0;

   // create primary header
   _create_primary_header(pu8_Buffer, SpacePacket::TM, SpacePacket::Unsegmented, 0x7ff, u16_SequenceCount,
       false, (uint32_t)u16_TargetPacketSize-PrimaryHdrSize);

   memset((char*)&pu8_Buffer[PrimaryHdrSize], 0xff, u16_TargetPacketSize-PrimaryHdrSize);

   return u16_TargetPacketSize;
}                               



int32_t SpacePacket::_create_primary_header(uint8_t *pu8_Buffer,
                 const PacketType e_PacketType, const SequenceFlags e_SequenceFlags, const uint16_t u16_APID, const uint16_t u16_SequenceCount, const bool b_SecHeader, 
                 const uint32_t u32_PacketDataLength)
{
   pu8_Buffer[0] = (uint8_t)(((SpPacketVersion&0x7)<<5) | ((e_PacketType&0x1)<<4) | ((b_SecHeader?1:0)<<3) | ((u16_APID>>8)&0x7));
   pu8_Buffer[1] = (uint8_t)(u16_APID&0xff);
   pu8_Buffer[2] = (uint8_t)(((e_SequenceFlags&0x3)<<6) | ((u16_SequenceCount>>8)&0x3f));
   pu8_Buffer[3] = (uint8_t)(u16_SequenceCount&0xff);
   pu8_Buffer[4] = (uint8_t)((u32_PacketDataLength-1)>>8);
   pu8_Buffer[5] = (uint8_t)((u32_PacketDataLength-1)&0xff);

   return 0;
}                               






int32_t SpacePacket::reset(void)
{
  if((mu32_Index>0) && (mu16_SyncErrorCount<0xffff))
    mu16_SyncErrorCount++;
  mu32_Index = 0;
  mb_Overflow = false;
  return 0;
}



int32_t SpacePacket::process(const uint8_t *pu8_Buffer, const uint32_t u32_BufferSize)
{
  if((u32_BufferSize==0) || !pu8_Buffer)
    return -1;

  for(uint32_t i=0; i<u32_BufferSize; i++)
  {
    switch(mu32_Index)
    {
      case 0:
        mu8_PacketVersionNumber = (uint8_t)((pu8_Buffer[i]&0xe0)>>5);
        me_PacketType = (PacketType)((pu8_Buffer[i]&0x10)>>4);
        mb_SecHdrFlag = (((pu8_Buffer[i]&0x08)>>3)==1)?true:false;
        mu16_APID = (uint16_t)((pu8_Buffer[i]&0x07)<<8);
        break; 
      case 1:
        mu16_APID |= (uint16_t)pu8_Buffer[i];
        break; 
      case 2:
        me_SequenceFlags = (SequenceFlags)((pu8_Buffer[i]&0xc0)>>6);;
        mu16_PacketSequenceCount = (uint16_t)((pu8_Buffer[i]&0x003F)<<8);
        break; 
      case 3:
        mu16_PacketSequenceCount |= (uint16_t)(pu8_Buffer[i]);
        break; 
      case 4:
        mu16_PacketDataLength = (uint16_t)(pu8_Buffer[i]<<8);
        break; 
      case 5:
        mu16_PacketDataLength |= (uint16_t)(pu8_Buffer[i]);
        break; 
      default:
        if(mu32_Index-SP_HEADER_SIZE<SP_MAX_DATA_SIZE) 
        {
          au8_PacketData[mu32_Index-PrimaryHdrSize]=pu8_Buffer[i];
        }
        else
        {
          if(!mb_Overflow && mu16_OverflowErrorCount<0xffff)
            mu16_OverflowErrorCount++;
          mb_Overflow = true;
        }
        break;
    }
    mu32_Index++;
    if((mu32_Index>=SP_HEADER_SIZE) && (mu32_Index>=(SP_HEADER_SIZE+mu16_PacketDataLength+1UL)))
    {
      if(NULL!=mp_SpCallback)
        mp_SpCallback(mp_SpContext, me_PacketType, me_SequenceFlags, mu16_APID, mu16_PacketSequenceCount, mb_SecHdrFlag, au8_PacketData, mu16_PacketDataLength+1);
      mu32_Index = 0;
      mb_Overflow = false;
    }
  } 
  return 0;
}


uint16_t SpacePacket::getSyncErrorCount(void)
{
  return mu16_SyncErrorCount;
}


uint16_t SpacePacket::getOverflowErrorCount(void)
{
  return mu16_OverflowErrorCount;
}



void SpacePacket::clearErrorCounters(void)
{
  mu16_SyncErrorCount=0;
  mu16_OverflowErrorCount=0;
  return;
}



}

